from functools import cached_property
from typing import TYPE_CHECKING

from followthemoney.model import registry
from nomenklatura.statement import make_statement_table
from sqlalchemy import (
    NUMERIC,
    BooleanClauseList,
    Column,
    MetaData,
    Select,
    and_,
    desc,
    func,
    or_,
    select,
    text,
    union_all,
)

from ftmq.enums import Comparators, PropertyTypes
from ftmq.exceptions import ValidationError
from ftmq.filters import F

if TYPE_CHECKING:
    from ftmq.query import Q


class Sql:
    COMPARATORS = {
        Comparators["eq"]: "__eq__",
        Comparators["not"]: "__ne__",
        Comparators["in"]: "in_",
        Comparators.null: "is_",
        Comparators.gt: "__gt__",
        Comparators.gte: "__ge__",
        Comparators.lt: "__lt__",
        Comparators.lte: "__le__",
    }

    def __init__(self, q: "Q") -> None:
        self.q = q
        self.metadata = MetaData()
        self.table = make_statement_table(self.metadata)

    def get_expression(self, column: Column, f: F):
        op = self.COMPARATORS.get(str(f.comparator), str(f.comparator))
        op = getattr(column, op)
        return op(f.value)

    @cached_property
    def clause(self) -> BooleanClauseList:
        clauses = []
        if self.q.datasets:
            clauses.append(
                or_(
                    self.get_expression(self.table.c.dataset, f)
                    for f in self.q.datasets
                )
            )
        if self.q.schemata:
            clauses.append(
                or_(
                    self.get_expression(self.table.c.schema, f) for f in self.q.schemata
                )
            )
        if self.q.reversed:
            rclause = or_(
                and_(
                    self.table.c.prop_type == str(registry.entity),
                    self.get_expression(self.table.c.value, f),
                )
                for f in self.q.reversed
            )
            rq = select(self.table.c.canonical_id.distinct()).where(
                and_(rclause, *clauses)
            )
            clauses.append(self.table.c.canonical_id.in_(rq))
        if self.q.properties:
            clauses.append(
                or_(
                    and_(
                        self.table.c.prop == f.key,
                        self.get_expression(self.table.c.value, f),
                    )
                    for f in self.q.properties
                )
            )
        return and_(*clauses)

    @cached_property
    def canonical_ids(self) -> Select:
        q = select(self.table.c.canonical_id.distinct()).where(self.clause)
        if self.q.sort is None:
            q = q.limit(self.q.limit).offset(self.q.offset)
        return q

    @cached_property
    def all_canonical_ids(self) -> Select:
        return self.canonical_ids.limit(None).offset(None)

    @cached_property
    def _unsorted_statements(self) -> Select:
        return (
            select(self.table)
            .where(self.table.c.canonical_id.in_(self.canonical_ids))
            .order_by(self.table.c.canonical_id)
        )

    @cached_property
    def _sorted_statements(self) -> Select:
        if self.q.sort:
            if len(self.q.sort.values) > 1:
                raise ValidationError(
                    f"Multi-valued sort not supported for `{self.__class__.__name__}`"
                )
            prop = self.q.sort.values[0]
            value = self.table.c.value
            if PropertyTypes[prop].value == registry.number:
                value = func.cast(self.table.c.value, NUMERIC)
            inner = (
                select(
                    self.table.c.canonical_id,
                    func.group_concat(value).label("sortable_value"),
                )
                .where(
                    and_(
                        self.table.c.prop == prop,
                        self.table.c.canonical_id.in_(self.canonical_ids),
                    )
                )
                .group_by(self.table.c.canonical_id)
                .limit(self.q.limit)
                .offset(self.q.offset)
            )

            order_by = "sortable_value"
            if not self.q.sort.ascending:
                order_by = desc(order_by)
            order_by = [order_by, self.table.c.canonical_id]

            inner = inner.order_by(*order_by)

            return select(
                self.table.join(
                    inner, self.table.c.canonical_id == inner.c.canonical_id
                )
            ).order_by(*order_by)

    @cached_property
    def statements(self) -> Select:
        if self.q.sort:
            return self._sorted_statements
        return self._unsorted_statements

    @cached_property
    def count(self) -> Select:
        return (
            select(func.count(self.table.c.canonical_id.distinct()))
            .select_from(self.table)
            .where(self.clause)
        )

    @cached_property
    def datasets(self) -> Select:
        return (
            select(
                self.table.c.dataset, func.count(self.table.c.canonical_id.distinct())
            )
            .where(self.clause)
            .group_by(self.table.c.dataset)
        )

    @cached_property
    def schemata(self) -> Select:
        return (
            select(
                self.table.c.schema, func.count(self.table.c.canonical_id.distinct())
            )
            .where(self.clause)
            .group_by(self.table.c.schema)
        )

    @cached_property
    def dates(self) -> Select:
        return select(
            func.min(self.table.c.value),
            func.max(self.table.c.value),
        ).where(
            self.table.c.prop_type == "date",
            self.table.c.canonical_id.in_(self.all_canonical_ids),
        )

    @cached_property
    def countries(self) -> Select:
        return (
            select(
                self.table.c.value,
                func.count(self.table.c.canonical_id.distinct()),
            )
            .where(
                self.table.c.prop_type == "country",
                self.table.c.canonical_id.in_(self.all_canonical_ids),
            )
            .group_by(self.table.c.value)
        )

    @cached_property
    def aggregations(self) -> Select:
        qs = []
        for agg in self.q.aggregations:
            qs.append(
                select(
                    text(f"'{agg.prop}'"),
                    text(f"'{agg.func}'"),
                    getattr(func, agg.func)(self.table.c.value),
                ).where(
                    self.table.c.prop == agg.prop,
                    self.table.c.canonical_id.in_(self.all_canonical_ids),
                )
            )
        return union_all(*qs)



import re
from collections.abc import Generator, Iterable
from functools import cache
from typing import Any

import pycountry
from banal import ensure_list
from nomenklatura.dataset import DataCatalog, Dataset, DefaultDataset
from nomenklatura.entity import CE, CompositeEntity
from nomenklatura.statement import Statement

from ftmq.enums import Comparators
from ftmq.exceptions import ValidationError
from ftmq.types import SGenerator


@cache
def make_dataset(name: str) -> Dataset:
    catalog = DataCatalog(
        Dataset, {"datasets": [{"name": name, "title": name.title()}]}
    )
    return catalog.get(name)


def parse_comparator(key: str) -> tuple[str, Comparators]:
    key, *comparator = key.split("__", 1)
    if comparator:
        comparator = Comparators[comparator[0]]
    else:
        comparator = Comparators["eq"]
    return key, comparator


def parse_unknown_filters(
    filters: tuple[str],
) -> Generator[tuple[str, str, str], None, None]:
    filters = (f for f in filters)
    for prop in filters:
        prop = prop.lstrip("-")
        if "=" in prop:  # 'country=de'
            prop, value = prop.split("=")
        else:  # ("country", "de"):
            value = next(filters)

        prop, *op = prop.split("__")
        yield prop, value, op[0] if op else None


def make_proxy(data: dict[str, Any], dataset: str | Dataset | None = None) -> CE:
    datasets = ensure_list(data.pop("datasets", None))
    if dataset is not None:
        if isinstance(dataset, str):
            dataset = dataset = make_dataset(dataset)
        datasets.append(dataset.name)
    elif datasets:
        dataset = datasets[0]
        dataset = make_dataset(dataset)
    else:
        dataset = DefaultDataset
    proxy = CompositeEntity(dataset, data)
    if len(datasets) > 1:
        if proxy.id is None:
            raise ValidationError("Entity has no ID.")
        statements = get_statements(proxy, *datasets)
        return CompositeEntity.from_statements(dataset, statements)
    return proxy


def get_statements(proxy: CE, *datasets: Iterable[str]) -> SGenerator:
    datasets = datasets or ["default"]
    for dataset in datasets:
        # FIXME
        for stmt in Statement.from_entity(proxy, dataset):
            stmt = stmt.to_dict()
            stmt["target"] = stmt.get("target") or False
            stmt["external"] = stmt.get("external") or False
            stmt = Statement.from_dict(stmt)
            yield stmt


@cache
def get_country_name(alpha2: str) -> str:
    try:
        country = pycountry.countries.get(alpha_2=alpha2.lower())
        return country.name
    except (LookupError, AttributeError):
        return alpha2


NUMERIC_US = re.compile(r"^-?\d+(?:,\d{3})*(?:\.\d+)?$")
NUMERIC_DE = re.compile(r"^-?\d+(?:\.\d{3})*(?:,\d+)?$")


def to_numeric(value: str) -> float | int | None:
    value = str(value).strip()
    try:
        value = float(value)
        if int(value) == value:
            return int(value)
        return value
    except ValueError:
        if re.match(NUMERIC_US, value):
            return to_numeric(value.replace(",", ""))
        if re.match(NUMERIC_DE, value):
            return to_numeric(value.replace(".", "").replace(",", "."))



import statistics
from collections import defaultdict
from functools import cache
from typing import Any, Iterable, TypeAlias

from banal import ensure_list
from followthemoney.schema import Schema
from followthemoney.types import registry
from pydantic import BaseModel

from ftmq.enums import Aggregations, Properties
from ftmq.types import CE, CEGenerator
from ftmq.util import to_numeric

Value: TypeAlias = int | float | str


@cache
def get_is_numeric(schema: Schema, prop: str) -> bool:
    prop = schema.get(prop)
    if prop is not None:
        return prop.type == registry.number
    return False


class Aggregation(BaseModel):
    prop: Properties
    func: Aggregations
    values: list[int | float] = []
    value: Value | None = None

    def __hash__(self) -> int:
        return hash((self.prop, self.func))

    def __eq__(self, other: Any) -> bool:
        return hash(self) == hash(other)

    def get_value(self) -> Value | None:
        if self.func == "min":
            return min(self.values)
        if self.func == "max":
            return max(self.values)
        if self.func == "sum":
            return sum(self.values)
        if self.func == "avg":
            return statistics.mean(self.values)

    def collect(self, proxy: CE) -> CE:
        is_numeric = get_is_numeric(proxy.schema, self.prop)
        for value in proxy.get(self.prop, quiet=True):
            if is_numeric:
                value = to_numeric(value)
            if value is not None:
                self.values.append(value)
        return proxy

    def apply(self, proxies: CEGenerator) -> CEGenerator:
        for proxy in proxies:
            yield self.collect(proxy)
        self.__exit__()

    def __enter__(self) -> "Aggregation":
        return self

    def __exit__(self, *args, **kwargs) -> None:
        self.value = self.get_value()

    def dict(self, *args, **kwargs) -> dict[str, Any]:
        self.__exit__()
        return super().dict(*args, **kwargs)


AggregatorResult: TypeAlias = dict[Aggregations, dict[Properties, Value]]


class Aggregator(BaseModel):
    aggregations: list[Aggregation] = []
    result: AggregatorResult = defaultdict(dict)

    def __enter__(self) -> "Aggregator":
        return self

    def __exit__(self, *args, **kwargs) -> None:
        for agg in self.aggregations:
            self.result[str(agg.func)][str(agg.prop)] = agg.value

    def apply(self, proxies: CEGenerator) -> CEGenerator:
        for agg in self.aggregations:
            proxies = agg.apply(proxies)
        yield from proxies
        self.__exit__()

    @classmethod
    def from_dict(cls, data: dict[Aggregations, Iterable[Properties]]) -> "Aggregator":
        return cls(
            aggregations=[
                Aggregation(prop=p, func=agg)
                for agg, props in data.items()
                for p in ensure_list(props)
            ]
        )

    def to_dict(self) -> dict[str, set[str]]:
        data = defaultdict(set)
        for agg in self.aggregations:
            data[str(agg.func)].add(str(agg.prop))
        return dict(data)
